// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "WCSinfo.h"
#include <errstr.h>

#include "param.h"
#include "ParamInfoSet.h"
#include "Region.h"

#include <cmath>


using namespace std;

//WCSinfo::WCSinfo(DataSet *ds)
//: _ds(ds)
WCSinfo::WCSinfo(Array * block, std::string coordtype, std::string shape_bkg_1strow, std::vector<double> src_and_bkg_regions)
: _array(block), _coordtype(coordtype), _shape_bkg_1strow(shape_bkg_1strow), _src_and_bkg_regions(src_and_bkg_regions)

{
	if (!checkWCS())
		errstr<<"The input array does not contain any WCS information. "
						"The displayed source positions may therefore be incorrect."
					  <<Sas::warning("invalidWCS");

	readWCS();

	convertWorldtoPixel();

}

bool WCSinfo::checkWCS()
{
	//Array *array = _ds->array(0);
	if (_array->hasAttribute("CDELT1") && _array->hasAttribute("CDELT2")) {
			string ctype1 = _array->attribute("CTYPE1")->asString();
			string ctype2 = _array->attribute("CTYPE2")->asString();
			if (ctype1.find("RA") != 0 || ctype2.find("DEC") != 0)
				errstr
				  <<"The WCS axes types ["
				  <<ctype1.c_str()
				  <<","
				  <<ctype2.c_str()
				  <<"] do not correspond to an equatorial system projection.  "
					"The displayed source positions may therefore be incorrect."
				  << Sas::warning("invalidWCS");
		}

	return true;
}

void WCSinfo::readWCS()
{
		// Read attributes from  array
    	//
    	//WCSConvertor wcs(_array);
    	//_xType = wcs.xType();
    	//_yType = wcs.yType();
    	//_physicalRefPix = wcs.physicalRefPixel();
    	//_physicalRefCoords = wcs.physicalRefCoords();
    	//_xPixelSizeWorld = wcs.xWidth();			// CDELT1
    	//_yPixelSizeWorld = wcs.yWidth();			// CDELT2
    	////cout << "		physicalRefCoords    : " << test.x() <<", "<< test.y() << endl;

	string alt;
	if (_coordtype == "ALT")
		alt = "L";		// PENDING to include alt = ""

    	  if (
    	      _array->hasAttribute("CTYPE1"+alt) &&
			  _array->hasAttribute("CRPIX1"+alt) &&
			  _array->hasAttribute("CRVAL1"+alt) &&
			  //_array->hasAttribute("CDELT1"+alt) &&
			  _array->hasAttribute("CTYPE2"+alt) &&
			  _array->hasAttribute("CRPIX2"+alt) &&
			  _array->hasAttribute("CRVAL2"+alt) )
			  //_array->hasAttribute("CDELT2"+alt) )
    	    {
    	      _ctype1 = _array->attribute("CTYPE1"+alt)->asString();

    	      _crpix1 = _array->attribute("CRPIX1"+alt)->asReal();
    	      _crval1 = _array->attribute("CRVAL1"+alt)->asReal();
    	      //_cdelt1 = _array->attribute("CDELT1"+alt)->asReal();

    	      _ctype2 = _array->attribute("CTYPE2"+alt)->asString();

    	      _crpix2 = _array->attribute("CRPIX2"+alt)->asReal();
    	      _crval2 = _array->attribute("CRVAL2"+alt)->asReal();
    	      //_cdelt2 = _array->attribute("CDELT2"+alt)->asReal();
    	    } else {
    	    	errstr << "Input image not valid. Mandatory WCS keywords not found (CTYPEn, CRVALn, CRPIXn, etc). Exiting."<< Sas::error("invalidImageType");
    	    }

    	  if ( _array->hasAttribute("CD1_1") ){
    		  _xMatrixOnAxis  = _array->attribute("CD1_1")->asReal();
    		  _xMatrixOffAxis = _array->attribute("CD2_1")->asReal();
    		  _yMatrixOffAxis = _array->attribute("CD1_2")->asReal();
    		  _yMatrixOnAxis  = _array->attribute("CD2_2")->asReal();
    	  } else {
    		  if (_array->hasAttribute("CDELT1"+alt) && _array->hasAttribute("CDELT2"+alt)){
    			  _cdelt1 = _array->attribute("CDELT1"+alt)->asReal();
    			  _cdelt2 = _array->attribute("CDELT2"+alt)->asReal();
    			  _xPixelSizeWorld = _cdelt1;
    			  _yPixelSizeWorld = _cdelt2;
    		  } else {
    			  errstr << "Input image not valid. Mandatory WCS keywords not found (CDELTn). Exiting."<< Sas::error("invalidImageType");
    		  }

    		  if (_array->hasAttribute("CROTA2")){
    		      		  _rotAngle = _array->attribute("CROTA2")->asReal();
    		  } else {_rotAngle = 0.0; }


    		  //WCSinfo convert;
    		  //_MatrixOnOffAxis = convert.convertOldToNew();
    		  convertOldToNew();
    	  }

    	  if (_array->hasAttribute("CUNIT1")){
    		  _XworldUnit = _array->attribute("CUNIT1")->asString();
    	  }
    	  if (_array->hasAttribute("CUNIT2")){
    	      _YworldUnit = _array->attribute("CUNIT2")->asString();
    	  }

}


void WCSinfo::convertWorldtoPixel()
{
    // Initializing parameters. Taken from 'implot'
    //vector<double> wcsThingie;		//  = (/ 0, 1, 2, 0/);
    struct wcsprm* wcs = new struct wcsprm;
    wcs->flag = 0;

    struct celprm* cel = new struct celprm;
      cel->flag = 0;
      cel->ref[0] = _crval1;						// wcsRefWorld.x();
      cel->ref[1] = _crval2;						// wcsRefWorld.y();
      cel->ref[2] = cel->ref[3] = 999.0;

    struct prjprm* prj = new struct prjprm;
        prj->flag = 0;
        prj->r0 = 0;
        for(unsigned i=0; i<10; i++) prj->p[i] = 0;

    //double pixelSizeXWorld = -sqrt(MatrixOnOffAxis[0] * MatrixOnOffAxis[0] + MatrixOnOffAxis[1] * MatrixOnOffAxis[1]);
    //double pixelSizeYWorld =  sqrt(MatrixOnOffAxis[2] * MatrixOnOffAxis[2] + MatrixOnOffAxis[3] * MatrixOnOffAxis[3]);
    double pixelSizeXWorld = _cdelt1;							// PixelSizeWorld.x();
    double pixelSizeYWorld = _cdelt2;							// PixelSizeWorld.y();
    //cout << endl;
    //cout << "		pixelSizeXWorld		: " << pixelSizeXWorld << endl;
    //cout << "		pixelSizeYWorld		: " << pixelSizeYWorld << endl;
    //cout << endl;

    double crpix[2], m[4], cdelt[2];
      crpix[0] = _crpix1;										// wcsRefPixels.x();		// xRefPixels
      crpix[1] = _crpix2;										// wcsRefPixels.y();		// yRefPixels
      m[0] = m[3] = 1;
      m[1] = m[2] = 0;
      cdelt[0] = pixelSizeXWorld;		// xPixelSizeWorld
      cdelt[1] = pixelSizeYWorld;		// yPixelSizeWorld

    struct linprm* lin = new struct linprm;
        lin->flag = 0;
        lin->naxis = 2;
        lin->crpix = crpix;
        lin->pc = m;
        lin->cdelt = cdelt;

    double crval[2];						// crval vector for 'wcsfwd'
        crval[0] = _crval1;											// wcsRefWorld.x();
        crval[1] = _crval2;											// wcsRefWorld.y();

    char ctype[2][9];						// ctype vector for 'wcsfwd'
        //ctype[0] = _ctype1;						// strcpy(ctype[0], ctype1.c_str());
        //ctype[1] = _ctype2;						// strcpy(ctype[1], ctype2.c_str());
        strcpy(ctype[0], _ctype1.c_str());
        strcpy(ctype[1], _ctype2.c_str());


    //double phi, theta, imgcrd[2], pixcrd[2], imgcrd_bkg[2], pixcrd_bkg[2];
    double phi, theta, imgcrd[2], pixcrd[2];

    // Check wcsfwd parameters:
    //cout << "		ctype		: " << ctype[0] << ", " << ctype[1] << endl;
    //cout << "		phi,theta	: " << phi << ", " << theta << endl;
    //cout << endl;

    //cout << "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - " << endl;
    double world[2];
    unsigned long regions_size = _src_and_bkg_regions.size();
    //cout << "	DEBUG =====> regions_size, regions_size/3 : " << regions_size << ", " << regions_size/3 << endl;
    for (unsigned long i = 0 ; i < (regions_size/3) ; ++i){

    	if ( _shape_bkg_1strow == "CIRCLE" ){
    	world[0] = _src_and_bkg_regions[0+(i*3)];
    	world[1] = _src_and_bkg_regions[1+(i*3)];
    	} else if ( _shape_bkg_1strow == "ANNULUS" ){
    		if (i < 2){
    			world[0] = _src_and_bkg_regions[0+(i*3)];
    			world[1] = _src_and_bkg_regions[1+(i*3)];
    		} else {
    			world[0] = _src_and_bkg_regions[1+0+(i*3)];
    			world[1] = _src_and_bkg_regions[1+1+(i*3)];
    		}
    	}
    	//for (unsigned long j = 0; j < 2; j++){
    	//	cout << "-- _src_and_bkg_regions, i, j : " << i << ", "<< j << ", " << _src_and_bkg_regions[j+(i*3)] << endl;
    	//}
    	//cout << "	world : " << world[0] << ", "<< world[1] << endl;

    	// Convert X/Y Sky to pixel x,y
    	int result = wcsfwd(ctype, wcs, world, crval, cel, &phi, &theta, prj, imgcrd, lin, pixcrd);
    	if(result != 0)
    	       errstr << "Cannot transform wcs coordinates (" << world[0] << "," << world[1]
    	   	   << ") to pixel coordinates.  wcsfwd error " << result << Sas::error("WcsTransform");

    	//std::vector<double> pixVec;
    	    _pixVec.push_back(pixcrd[0]);
    	    _pixVec.push_back(pixcrd[1]);
    	    //for (unsigned long i = 0 ; i < _pixVec.size() ; ++i){
    	    //cout << "		pixVec		: " << _pixVec[i] << endl;
    	    //}

    }
	    delete wcs;
	    delete cel;
	    free(lin->piximg);
	    free(lin->imgpix);
	    delete lin;

    //for (unsigned long i = 0 ; i < (_pixVec.size()/2) ; ++i){
    //	cout << "		pixVec		: " << _pixVec[0+(i*2)] << ", " << _pixVec[1+(i*2)] << endl;
    //}

    //cout << "- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - " << endl;


}


void WCSinfo::convertOldToNew()
{
/*
! Taken from 'ssclib/src/wcs_aux.f90'
! Makes use of equation (189) from Calabretta & Greisen A&A 395 1077
! (2002), ie
!
!	(CD1_1 CD1_2)   (CDELT1*cos[CROTA2]  -CDELT2*sin[CROTA2])
!	(           ) = (                                       ).
!	(CD2_1 CD2_2)   (CDELT1*sin[CROTA2]   CDELT2*cos[CROTA2])
*/
	_xMatrixOnAxis  =  _xPixelSizeWorld * cos(_rotAngle);
	_xMatrixOffAxis =  _xPixelSizeWorld * sin(_rotAngle);
	_yMatrixOffAxis = -_yPixelSizeWorld * sin(_rotAngle);
	_yMatrixOnAxis  =  _yPixelSizeWorld * cos(_rotAngle);

		_MatrixOnOffAxis.push_back(_xMatrixOnAxis);
		_MatrixOnOffAxis.push_back(_xMatrixOffAxis);
		_MatrixOnOffAxis.push_back(_yMatrixOnAxis);
		_MatrixOnOffAxis.push_back(_yMatrixOffAxis);

	//cout <<endl;
	//cout << "		_xPixelSizeWorld, _yPixelSizeWorld : " << _xPixelSizeWorld <<", "<< _yPixelSizeWorld << endl;
	//cout <<endl;
	//cout << "		_xMatrixOnAxis, _xMatrixOffAxis : " << _xMatrixOnAxis <<", "<< _xMatrixOffAxis << endl;
	//cout << "		_yMatrixOnAxis, _yMatrixOffAxis : " << _yMatrixOnAxis <<", "<< _yMatrixOffAxis << endl;
	//cout <<endl;
}


string WCSinfo::getxType()
//const char * WCSinfo::getxType()
{
  return _ctype1;
}
string WCSinfo::getyType()
//const char * WCSinfo::getyType()
{
  return _ctype2;
}

string WCSinfo::getXUnit()
{
  return _XworldUnit;
}
string WCSinfo::getYUnit()
{
  return _YworldUnit;
}

CoordPair WCSinfo::getRefPixels()
{
	return CoordPair (_crpix1, _crpix2);
}

CoordPair WCSinfo::getRefCoords()
{
	return CoordPair (_crval1, _crval2);
}

CoordPair WCSinfo::getPixelSizeWorld()
{
	return CoordPair (_cdelt1, _cdelt2);
}

real64 WCSinfo::getrotAngle()
{
	return _rotAngle;
}

std::vector<real64> WCSinfo::getMatrixOnOffAxis()
{
	return _MatrixOnOffAxis;
}

std::vector<double> WCSinfo::getPixVec()
{
	return _pixVec;
}
