// ESA (C) 2000-2018 
//
// This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
//
//    SAS is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    SAS is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
#include "loadImage.h"

#include <Tracer.h>
#include <errstr.h>
#include "CalServer.h"
#include "Dal.h"
#include <numeric>

//#include <cmath>
//#include <algorithm>
//#include <limits>

using namespace std;

loadImage::loadImage(string imagefile, string regcoordtype, string shape_bkg_1strow, vector<double> src_and_bkg_regions, int backscale, double total_spec_counts, double bkg_spec_exposure)
: _imagefile(imagefile), _regcoordtype(regcoordtype), _shape_bkg_1strow(shape_bkg_1strow), _src_and_bkg_regions(src_and_bkg_regions), _backscale(backscale), _total_spec_counts(total_spec_counts), _bkg_spec_exposure(bkg_spec_exposure)
{
	calImage();
}

void loadImage::calImage()
{
	DataSet	*imageSet = dataSetServer->open(_imagefile,Dal::Read);
	// Get instrument and obs. mode from image file
	//std::string datamode;
	//std::string submode;

		if (imageSet->hasAttribute("INSTRUME"))
			_instr = imageSet->attribute("INSTRUME")->asString();

		if (imageSet->hasAttribute("EXPIDSTR"))
			_expid = imageSet->attribute("EXPIDSTR")->asString();

		if (imageSet->hasAttribute("OBS_ID") )
			_obs_id = imageSet->attribute("OBS_ID")->asString();

		if (imageSet->hasAttribute("DATAMODE"))
			_datamode = imageSet->attribute("DATAMODE")->asString();

		if (imageSet->hasAttribute("SUBMODE") )
			_submode = imageSet->attribute("SUBMODE")->asString();

		if (imageSet->hasAttribute("FILTER") )
			_filter = imageSet->attribute("FILTER")->asString();
		if (imageSet->hasAttribute("OBJECT") )
			_object = imageSet->attribute("OBJECT")->asString();
		if (imageSet->hasAttribute("OBSERVER") )
			_observer = imageSet->attribute("OBSERVER")->asString();
		if (imageSet->hasAttribute("DATE-OBS") )
			_date_obs = imageSet->attribute("DATE-OBS")->asString();
		//if (imageSet->hasAttribute("CTYPE1")){
		//	cout << "		EXPOSURE from imageSet YES" << endl;
		//	//_exposure = imageSet->attribute("EXPOSURE")->asReal();
		//} //else {_exposure = 0;}


	// Check image input
	if (_datamode != "IMAGING"){
		errstr << "Input image not valid. Exiting."<< Sas::error("invalidImageType");
	}

	//cout << endl;
	//cout << "Processing input image : " << _imagefile << endl;
	//cout << "		Instrument : " << _instr << endl;
	//cout << "		Obsid      : " << _obs_id << endl;
	//cout << "		Datamode   : " << _datamode << endl;
	//cout << "		Submode    : " << _submode << endl;
	//cout << endl;


	Array *imageArray = imageSet->array(0);			// 0 = PRIMARY extension
	unsigned int ndim = imageArray->dimensions();
	vector <unsigned long> size = imageArray->size();
			int nx = size[1];
			int ny = size[0];
			//cout << "		ndim   : " << ndim << endl;
			//cout << "		size   : " << size[0] << ", " << size[1] << endl;
			if( ndim != 2 || size[0] < 2 || size[1] < 2 ){
					dataSetServer->close(imageSet);
				   Sas::error("WrongImage");
				}
			_resSize.push_back(nx);
			_resSize.push_back(ny);

		// For some reason EXPOSURE keyword has to be read from imageArray, but not imageSet
		if (imageArray->hasAttribute("EXPOSURE")){
			_exposure = imageArray->attribute("EXPOSURE")->asReal();
		} else {_exposure = 0;}
		//cout << "		EXPOSURE from imageSet   : " << _exposure << endl;
		//cout << "		EXPOSURE from Bkg Spec   : " << _bkg_spec_exposure << endl;


	ArrayData *arrayData = imageArray -> data();
	int32 * data = arrayData -> int32Data();

	///////////////////////////////////////////////////////////////////////////////////////////////
	// Read WCS information from the input image
	//
		string coordtype;
		if (_regcoordtype == "pos")				// regcoordtype = pos => X,Y => Alternate "ALT" reference WCS keywords: CRPIXnL, CRVALnL, etc
			coordtype = "ALT";
		else if (_regcoordtype == "eqpos")		// regcoordtype = eqpos => RA,Dec => Normal reference WCS keywords: CRPIXn, CRVALn, etc
			coordtype = "";						// It is not completed .TBD

	WCSinfo wcs(imageArray, coordtype, _shape_bkg_1strow, _src_and_bkg_regions);		// Read alternative attributes "ALT": CTYPE1/2L, CRPIX1/2L, etc
	//
	_ctype1 = wcs.getxType();
	_ctype2 = wcs.getyType();
	_refPixels = wcs.getRefPixels();
	_refCoords = wcs.getRefCoords();
	_PixelSizeWorld = wcs.getPixelSizeWorld();			// _PixelSizeWorld.x/y() = 80
	_XworldUnit = wcs.getXUnit();
	_YworldUnit = wcs.getYUnit();
	_rotAngle = wcs.getrotAngle();
	_MatrixOnOffAxis = wcs.getMatrixOnOffAxis();
	_pixVec = wcs.getPixVec();
	///////////////////////////////////////////////////////////////////////////////////////////////


	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	//
	// Source and Background regions
	double square_src[4];
		square_src[0] = _src_and_bkg_regions[0]/_PixelSizeWorld.x() - _src_and_bkg_regions[2]/_PixelSizeWorld.x();		// Src_X1 = Src_X0 - rad
		square_src[1] = _src_and_bkg_regions[0]/_PixelSizeWorld.x() + _src_and_bkg_regions[2]/_PixelSizeWorld.x();		// Src_X2 = Src_X0 + rad
		square_src[2] = _src_and_bkg_regions[1]/_PixelSizeWorld.x() - _src_and_bkg_regions[2]/_PixelSizeWorld.x();		// Src_Y1 = Src_Y0 - rad
		square_src[3] = _src_and_bkg_regions[1]/_PixelSizeWorld.x() + _src_and_bkg_regions[2]/_PixelSizeWorld.x();		// Src_Y2 = Src_Y0 + rad

	double square_bkg[4], square_bkg_area;
	double big_square_bkg[4];
	double small_square_bkg[4];
	//cout << "		=====> _shape_bkg_1strow  : " << _shape_bkg_1strow << endl;
	if ( _shape_bkg_1strow == "CIRCLE" ){
		square_bkg[0] = _src_and_bkg_regions[3]/_PixelSizeWorld.x() - _src_and_bkg_regions[5]/_PixelSizeWorld.x();		// Bkg_X1 = Bkg_X0 - rad_bkg
		square_bkg[1] = _src_and_bkg_regions[3]/_PixelSizeWorld.x() + _src_and_bkg_regions[5]/_PixelSizeWorld.x();		// Bkg_X2 = Bkg_X0 + rad_bkg
		square_bkg[2] = _src_and_bkg_regions[4]/_PixelSizeWorld.x() - _src_and_bkg_regions[5]/_PixelSizeWorld.x();		// Bkg_Y1 = Bkg_Y0 - rad_bkg
		square_bkg[3] = _src_and_bkg_regions[4]/_PixelSizeWorld.x() + _src_and_bkg_regions[5]/_PixelSizeWorld.x();		// Bkg_Y2 = Bkg_Y0 + rad_bkg
		square_bkg_area = ( abs(square_bkg[1] - square_bkg[0]) )*(abs(square_bkg[3] - square_bkg[2]));

	} else if ( _shape_bkg_1strow == "ANNULUS" ) {	// _shape_bkg_1strow == "ANNULUS"
		//cout << "		=====> _shape_bkg_1strow INSIDE : " << _shape_bkg_1strow << endl;
		big_square_bkg[0] = _src_and_bkg_regions[3]/_PixelSizeWorld.x() - _src_and_bkg_regions[6]/_PixelSizeWorld.x();		// Bkg_X1 = Bkg_X0 - rad2_bkg
		big_square_bkg[1] = _src_and_bkg_regions[3]/_PixelSizeWorld.x() + _src_and_bkg_regions[6]/_PixelSizeWorld.x();		// Bkg_X2 = Bkg_X0 + rad2_bkg
		big_square_bkg[2] = _src_and_bkg_regions[4]/_PixelSizeWorld.x() - _src_and_bkg_regions[6]/_PixelSizeWorld.x();		// Bkg_Y1 = Bkg_Y0 - rad2_bkg
		big_square_bkg[3] = _src_and_bkg_regions[4]/_PixelSizeWorld.x() + _src_and_bkg_regions[6]/_PixelSizeWorld.x();		// Bkg_Y2 = Bkg_Y0 + rad2_bkg

		small_square_bkg[0] = _src_and_bkg_regions[3]/_PixelSizeWorld.x() - _src_and_bkg_regions[5]/_PixelSizeWorld.x();		// Bkg_X1 = Bkg_X0 - rad1_bkg
		small_square_bkg[1] = _src_and_bkg_regions[3]/_PixelSizeWorld.x() + _src_and_bkg_regions[5]/_PixelSizeWorld.x();		// Bkg_X2 = Bkg_X0 + rad1_bkg
		small_square_bkg[2] = _src_and_bkg_regions[4]/_PixelSizeWorld.x() - _src_and_bkg_regions[5]/_PixelSizeWorld.x();		// Bkg_Y1 = Bkg_Y0 - rad1_bkg
		small_square_bkg[3] = _src_and_bkg_regions[4]/_PixelSizeWorld.x() + _src_and_bkg_regions[5]/_PixelSizeWorld.x();		// Bkg_Y2 = Bkg_Y0 + rad1_bkg
		cout << "		big_square_src    : " << big_square_bkg[0] << ", " << big_square_bkg[1] << ", "
										      << big_square_bkg[2] << ", " << big_square_bkg[3] << endl;
		cout << "		small_square_bkg    : " << small_square_bkg[0] << ", " << small_square_bkg[1] << ", "
										        << small_square_bkg[2] << ", " << small_square_bkg[3] << endl;
		cout << endl;
	}
		//cout << "		square_src    : " << square_src[0] << ", " << square_src[1] << ", "
		//								  << square_src[2] << ", " << square_src[3] << endl;
		//cout << "		square_bkg    : " << square_bkg[0] << ", " << square_bkg[1] << ", "
		//								  << square_bkg[2] << ", " << square_bkg[3] << endl;
		//cout << endl;
	////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


	   _image = new float[nx*ny];
	   std::vector<int> x_indeces;
	   std::vector<int> y_indeces;
	   std::vector<double> src_pixel_values_in_region;
	   std::vector<double> bkg_pixel_values_in_region;
	    for (int i = 0; i < nx; i++){
	                for (int j = 0; j < ny; j++){
	        	      _image[i + (j * nx)] = data[i + (j * nx)];	// _image population from 'data' 2D array

	        	      	  // Estimation of the part of the image with real data
	        	      	  if (_image[i + (j * nx)] > 0 ){	// pixel value > 0 ==> Estimation of the part of the image with real data
	        	      		x_indeces.push_back(i+1);		// 1st array position = 0. 1st pixel position = 1
	        	      		y_indeces.push_back(j+1);		// 1st array position = 0. 1st pixel position = 1
	        	      	  }

	        	      	  // Pixel values in the Source region
	        	      	  if (i+1 > square_src[0] && i+1 < square_src[1] && j+1 > square_src[2] && j+1 < square_src[3]){
	        	      		//cout << "		Image pixel value   : " << _image[i + (j * nx)] << endl;
	        	      		src_pixel_values_in_region.push_back(_image[i + (j * nx)]);
	        	      	  }
	        	      	  // Pixel values in the Background region
	        	      	  if ( _shape_bkg_1strow == "CIRCLE" ){
	        	      		  if (i+1 > square_bkg[0] && i+1 < square_bkg[1] && j+1 > square_bkg[2] && j+1 < square_bkg[3]){
	        	      		  //cout << "		Image pixel value   : " << _image[i + (j * nx)] << endl;
	        	      		  bkg_pixel_values_in_region.push_back(_image[i + (j * nx)]);}

	        	      	  } else if ( _shape_bkg_1strow == "ANNULUS" ){	// _shape_bkg_1strow == "ANNULUS"
	        	      		// Pixel into the big square, but around the small square
	        	      		if (
	        	      			(i+1 > big_square_bkg[0] && i+1 < small_square_bkg[0]) || (i+1 < big_square_bkg[1] && i+1 > small_square_bkg[1]) ||
								(j+1 > big_square_bkg[2] && j+1 < small_square_bkg[2]) || (j+1 < big_square_bkg[3] && j+1 > small_square_bkg[3])
	        	      		   )
							{
	        	      				//cout << "		Image pixel value bkg ANNULUS  : " << _image[i + (j * nx)] << endl;
	        	      		  bkg_pixel_values_in_region.push_back(_image[i + (j * nx)]);}
	        	      	  }

	                }
	        }


	   // Limits with data in image == clipBlankBorder == remove blank around the part of the image with data
	   std::vector<int>::const_iterator forminx =  std::min_element(x_indeces.begin(), x_indeces.end());
	   std::vector<int>::const_iterator formaxx =  std::max_element(x_indeces.begin(), x_indeces.end());
	   std::vector<int>::const_iterator forminy =  std::min_element(y_indeces.begin(), y_indeces.end());
	   std::vector<int>::const_iterator formaxy =  std::max_element(y_indeces.begin(), y_indeces.end());
	   _minx = *forminx;
	   _maxx = *formaxx;
	   _miny = *forminy;
	   _maxy = *formaxy;
	   _XY_limits.push_back(_minx);
	   _XY_limits.push_back(_maxx);
	   _XY_limits.push_back(_miny);
	   _XY_limits.push_back(_maxy);
	   //cout << "		minx, maxx, miny, maxy   : " << _minx << ", " << _maxx << " ; " << _miny << ", " << _maxy << endl;

	delete arrayData;
	dataSetServer->close(imageSet);


	double effMin;
	double effMax;
	//
	// Brightness cuts are automatic. Not configurable
	//
	double cdelt1L = _PixelSizeWorld.x();
	double total_cts_bkg = accumulate(bkg_pixel_values_in_region.begin(), bkg_pixel_values_in_region.end(), 0.0f);

		if (_total_spec_counts == 0){	// ==> "No SPECTRUM extension in the BKG Region file. Brightness cuts will be estimated from the image. "
			// effMin == Min cut == [Total(counts in bkg region)] / [BACKSCAL / 80 ^ 2] // CHANGE ALL 80 by CDELT1L !!!!!!

			effMin = (total_cts_bkg)/( square_bkg_area );		// Total Counts from pixels into Region

			//cout << "		square_bkg   : " << square_bkg[0] << ", " << square_bkg[1] << " ; " << square_bkg[2] << ", " << square_bkg[3] << endl;
			//cout << "		total_cts_bkg, square_bkg_area   : " << total_cts_bkg << ", " << square_bkg_area << endl;
			//cout << "		cdelt1L, effMin   : " << cdelt1L << ", " << effMin << endl;

		} else {						// ==> effMin calculated from total_spec_counts == total counts in BKG Spec

			if (_backscale == 0 ){
				effMin = 0.0;
			} else {
				// Exposure Time Scale (Image Exposure Time / Bkg Spec Exposure Time ) == ( _exposure / _bkg_spec_exposure ):
				double exposure_scale = _exposure / _bkg_spec_exposure;

				double effMin_reg = (total_cts_bkg)/( (_backscale)/(pow(cdelt1L,2)));							// Total Counts from pixels into BKG Region
				double effMin_spec = (_total_spec_counts)/( (_backscale)/(pow(cdelt1L,2))) * exposure_scale;	// Total Counts from BKG Spectrum
				//cout << "		effMin_reg, effMin_spec   : " << effMin_reg << ", " << effMin_spec << endl;
				std::vector<double> effMin_values;
				  effMin_values.push_back(effMin_reg);
				  effMin_values.push_back(effMin_spec);
				  std::vector<double>::const_iterator effMin_min =  std::min_element(effMin_values.begin(), effMin_values.end());
				  //cout << "		effMin_reg, effMin_spec, Min   : " << effMin_values[0] << ", " << effMin_values[1] << ", " << *effMin_min << endl;
				  //if (effMin_reg > effMin_spec){
				  //  cout << "			======> effMin_reg, effMin_spec   : " << effMin_reg << ", " << effMin_spec << endl;
				  //	  cout << endl;
				  //}

				effMin = *effMin_min;		// Select the minor between effMin_reg and effMin_spec
			}
		}

		// effMax == Max cut == Max pixel value into the Source region
		std::vector<double>::const_iterator src_max =  std::max_element(src_pixel_values_in_region.begin(), src_pixel_values_in_region.end());
		effMax = *src_max;
		// Min pixel value into the Source region. Only used if effMin > effMax
		std::vector<double>::const_iterator src_min =  std::min_element(src_pixel_values_in_region.begin(), src_pixel_values_in_region.end());
		//src_min_value = *src_min;

		//cout << "		effMin, effMax   : " << effMin << ", " << effMax << endl;
	//}

	if ( effMin > effMax ){
		//errstr << "The minimum and maximum image brightness truncation "
		//		"levels are inverted! Exiting."<< Sas::error("badZclipLevels");
		errstr<<"Min cut > Max image cut, so average Background level is higher than the Source level. Image cuts fixed only to source." <<Sas::message(Msg::App, Msg::Sparse);
		effMin = *src_min;
	}
    _res_effMinMax.push_back(effMin);
    _res_effMinMax.push_back(effMax);
}




float * loadImage::getImage()
{
	return _image;
}

vector<int> loadImage::getImageSize()
{
	return _resSize;
}

vector<double> loadImage::getImageeffMinMax()
{
	return _res_effMinMax;
}


vector<int> loadImage::getXY_limits()
{
	return _XY_limits;
}

string loadImage::getobs_id()
{
	return _obs_id;
}

string loadImage::getInstr()
{
	return _instr;
}

string loadImage::getDatamode()
{
	return _datamode;
}

string loadImage::getSubmode()
{
	return _submode;
}

string loadImage::getExpid()
{
	return _expid;
}

string loadImage::getFilter()
{
	return _filter;
}

string loadImage::getObject()
{
	return _object;
}

string loadImage::getObserver()
{
	return _observer;
}

string loadImage::getDate()
{
	return _date_obs;
}

double loadImage::getExp()
{
	return _exposure;
}





///////////////////////////////////////////
// Get WCS info from the input image
string loadImage::getxType()
//const char * loadImage::getxType()
{
	return _ctype1;
}
string loadImage::getyType()
//const char * loadImage::getyType()
{
	return _ctype2;
}

string loadImage::getXUnit()
{
  return _XworldUnit;
}
string loadImage::getYUnit()
{
  return _YworldUnit;
}

CoordPair loadImage::getRefPixels()
{
	return _refPixels;
}

CoordPair loadImage::getRefCoords()
{
	return _refCoords;
}

CoordPair loadImage::getPixelSizeWorld()
{
	return _PixelSizeWorld;
}

real64 loadImage::getrotAngle()
{
	return _rotAngle;
}

std::vector<real64> loadImage::getMatrixOnOffAxis()
{
	return _MatrixOnOffAxis;
}

std::vector<double> loadImage::getPixVec()
{
	return _pixVec;
}
///////////////////////////////////////////

